"""Utility functions that get things"""

import logging
from voluptuous import Schema
from elasticsearch8 import Elasticsearch
from elasticsearch8.exceptions import NotFoundError
from es_client.helpers.schemacheck import SchemaCheck
from es_client.helpers.utils import prune_nones
from curator.helpers.getters import get_repository, get_write_index
from curator.exceptions import (
    ConfigurationError,
    MissingArgument,
    RepositoryException,
    SearchableSnapshotException,
)
from curator.defaults.settings import (
    index_filtertypes,
    snapshot_actions,
    snapshot_filtertypes,
)
from curator.validators import actions, options
from curator.validators.filter_functions import validfilters
from curator.helpers.utils import report_failure


def has_lifecycle_name(idx_settings):
    """
    :param idx_settings: The settings for an index being tested
    :type idx_settings: dict

    :returns: ``True`` if a lifecycle name exists in settings, else ``False``
    :rtype: bool
    """
    if 'lifecycle' in idx_settings:
        if 'name' in idx_settings['lifecycle']:
            return True
    return False


def is_idx_partial(idx_settings):
    """
    :param idx_settings: The settings for an index being tested
    :type idx_settings: dict

    :returns: ``True`` if store.snapshot.partial exists in settings, else ``False``
    :rtype: bool
    """
    if 'store' in idx_settings:
        if 'snapshot' in idx_settings['store']:
            if 'partial' in idx_settings['store']['snapshot']:
                if idx_settings['store']['snapshot']['partial']:
                    return True
                # store.snapshot.partial exists but is False -- Not a frozen tier mount
                return False
            # store.snapshot exists, but partial isn't there --
            # Possibly a cold tier mount
            return False
        raise SearchableSnapshotException('Index not a mounted searchable snapshot')
    raise SearchableSnapshotException('Index not a mounted searchable snapshot')


def ilm_policy_check(client, alias):
    """Test if alias is associated with an ILM policy

    Calls :py:meth:`~.elasticsearch.client.IndicesClient.get_settings`

    :param client: A client connection object
    :param alias: The alias name

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type alias: str
    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    # alias = action_obj.options['name']
    write_index = get_write_index(client, alias)
    try:
        idx_settings = client.indices.get_settings(index=write_index)
        if 'name' in idx_settings[write_index]['settings']['index']['lifecycle']:
            # logger.info('Alias %s is associated with ILM policy.', alias)
            # logger.info('Skipping action %s because allow_ilm_indices is false.', idx)
            return True
    except KeyError:
        logger.debug('No ILM policies associated with %s', alias)
    return False


def repository_exists(client, repository=None):
    """
    Calls :py:meth:`~.elasticsearch.client.SnapshotClient.get_repository`

    :param client: A client connection object
    :param repository: The Elasticsearch snapshot repository to use

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type repository: str

    :returns: ``True`` if ``repository`` exists, else ``False``
    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    if not repository:
        raise MissingArgument('No value for "repository" provided')
    try:
        test_result = get_repository(client, repository)
        if repository in test_result:
            logger.debug("Repository %s exists.", repository)
            response = True
        else:
            logger.debug("Repository %s not found...", repository)
            response = False
    # pylint: disable=broad-except
    except Exception as err:
        logger.debug('Unable to find repository "%s": Error: %s', repository, err)
        response = False
    return response


def rollable_alias(client, alias):
    """
    Calls :py:meth:`~.elasticsearch.client.IndicesClient.get_alias`

    :param client: A client connection object
    :param alias: An Elasticsearch alias

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type alias: str


    :returns: ``True`` or ``False`` depending on whether ``alias`` is an alias that
        points to an index that can be used by the ``_rollover`` API.
    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    try:
        response = client.indices.get_alias(name=alias)
    except NotFoundError:
        logger.error('Alias "%s" not found.', alias)
        return False
    # Response should be like:
    # {'there_should_be_only_one': {'aliases': {'value of "alias" here': {}}}}
    # where 'there_should_be_only_one' is a single index name that ends in a number,
    # and 'value of "alias" here' reflects the value of the passed parameter, except
    # where the ``is_write_index`` setting makes it possible to have more than one
    # index associated with a rollover index
    for idx in response:
        if 'is_write_index' in response[idx]['aliases'][alias]:
            if response[idx]['aliases'][alias]['is_write_index']:
                return True
    # implied ``else``: If not ``is_write_index``, it has to fit the following criteria:
    if len(response) > 1:
        logger.error(
            '"alias" must only reference one index, but points to %s', response
        )
        return False
    index = list(response.keys())[0]
    rollable = False
    # In order for `rollable` to be True, the last 2 digits of the index
    # must be digits, or a hyphen followed by a digit.
    # NOTE: This is not a guarantee that the rest of the index name is
    # necessarily correctly formatted.
    if index[-2:][1].isdigit():
        if index[-2:][0].isdigit():
            rollable = True
        elif index[-2:][0] == '-':
            rollable = True
    return rollable


def snapshot_running(client):
    """
    Calls :py:meth:`~.elasticsearch.client.SnapshotClient.get_repository`

    Return ``True`` if a snapshot is in progress, and ``False`` if not

    :param client: A client connection object

    :type client: :py:class:`~.elasticsearch.Elasticsearch`

    :rtype: bool
    """
    try:
        status = client.snapshot.status()['snapshots']
    # pylint: disable=broad-except
    except Exception as exc:
        report_failure(exc)
    # We will only accept a positively identified False.  Anything else is
    # suspect. That's why this statement, rather than just ``return status``
    # pylint: disable=simplifiable-if-expression
    return False if not status else True


def validate_actions(data):
    """
    Validate the ``actions`` configuration dictionary, as imported from actions.yml,
    for example.

    :param data: The configuration dictionary

    :type data: dict

    :returns: The validated and sanitized configuration dictionary.
    :rtype: dict
    """
    # data is the ENTIRE schema...
    clean_config = {}
    # Let's break it down into smaller chunks...
    # First, let's make sure it has "actions" as a key, with a subdictionary
    root = SchemaCheck(data, actions.root(), 'Actions File', 'root').result()
    # We've passed the first step.  Now let's iterate over the actions...
    for action_id in root['actions']:
        # Now, let's ensure that the basic action structure is correct, with
        # the proper possibilities for 'action'
        action_dict = root['actions'][action_id]
        loc = f'Action ID "{action_id}"'
        valid_structure = SchemaCheck(
            action_dict, actions.structure(action_dict, loc), 'structure', loc
        ).result()
        # With the basic structure validated, now we extract the action name
        current_action = valid_structure['action']
        # And let's update the location with the action.
        loc = f'Action ID "{action_id}", action "{current_action}"'
        clean_options = SchemaCheck(
            prune_nones(valid_structure['options']),
            options.get_schema(current_action),
            'options',
            loc,
        ).result()
        clean_config[action_id] = {
            'action': current_action,
            'description': valid_structure['description'],
            'options': clean_options,
        }
        if current_action == 'alias':
            add_remove = {}
            for k in ['add', 'remove']:
                if k in valid_structure:
                    current_filters = SchemaCheck(
                        valid_structure[k]['filters'],
                        Schema(validfilters(current_action, location=loc)),
                        f'"{k}" filters',
                        f'{loc}, "filters"',
                    ).result()
                    add_remove.update(
                        {
                            k: {
                                'filters': SchemaCheck(
                                    current_filters,
                                    Schema(validfilters(current_action, location=loc)),
                                    'filters',
                                    f'{loc}, "{k}", "filters"',
                                ).result()
                            }
                        }
                    )
            # Add/Remove here
            clean_config[action_id].update(add_remove)
        elif current_action in ['cluster_routing', 'create_index', 'rollover']:
            # neither cluster_routing nor create_index should have filters
            pass
        else:  # Filters key only appears in non-alias actions
            valid_filters = SchemaCheck(
                valid_structure['filters'],
                Schema(validfilters(current_action, location=loc)),
                'filters',
                f'{loc}, "filters"',
            ).result()
            clean_filters = validate_filters(current_action, valid_filters)
            clean_config[action_id].update({'filters': clean_filters})
        # This is a special case for remote reindex
        if current_action == 'reindex':
            # Check only if populated with something.
            if 'remote_filters' in valid_structure['options']:
                valid_filters = SchemaCheck(
                    valid_structure['options']['remote_filters'],
                    Schema(validfilters(current_action, location=loc)),
                    'filters',
                    f'{loc}, "filters"',
                ).result()
                clean_remote_filters = validate_filters(current_action, valid_filters)
                clean_config[action_id]['options'].update(
                    {'remote_filters': clean_remote_filters}
                )

    # if we've gotten this far without any Exceptions raised, it's valid!
    return {'actions': clean_config}


def validate_filters(action, myfilters):
    """
    Validate that myfilters are appropriate for the action type, e.g. no
    index filters applied to a snapshot list.

    :param action: An action name
    :param myfilters: A list of filters to test.

    :type action: str
    :type myfilters: list

    :returns: Validated list of filters
    :rtype: list
    """
    # Define which set of filtertypes to use for testing
    if action in snapshot_actions():
        filtertypes = snapshot_filtertypes()
    else:
        filtertypes = index_filtertypes()
    for fil in myfilters:
        if fil['filtertype'] not in filtertypes:
            raise ConfigurationError(
                f"\"{fil['filtertype']}\" filtertype is not compatible with "
                f"action \"{action}\""
            )
    # If we get to this point, we're still valid.  Return the original list
    return myfilters


def verify_client_object(test):
    """
    :param test: The variable or object to test

    :type test: :py:class:`~.elasticsearch.Elasticsearch`

    :returns: ``True`` if ``test`` is a proper :py:class:`~.elasticsearch.Elasticsearch`
        client object and raise a :py:exc:`TypeError` exception if it is not.
    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    # Ignore mock type for testing
    if str(type(test)) == "<class 'unittest.mock.Mock'>":
        pass
    elif not isinstance(test, Elasticsearch):
        msg = f'Not a valid client object. Type: {type(test)} was passed'
        logger.error(msg)
        raise TypeError(msg)


def verify_index_list(test):
    """
    :param test: The variable or object to test

    :type test: :py:class:`~.curator.IndexList`

    :returns: ``None`` if ``test`` is a proper :py:class:`~.curator.indexlist.IndexList`
        object, else raise a :py:class:`TypeError` exception.
    :rtype: None
    """
    # It breaks if this import isn't local to this function:
    # ImportError: cannot import name 'IndexList' from partially initialized module
    # 'curator.indexlist' (most likely due to a circular import)
    # pylint: disable=import-outside-toplevel
    from curator.indexlist import IndexList

    logger = logging.getLogger(__name__)
    if not isinstance(test, IndexList):
        msg = f'Not a valid IndexList object. Type: {type(test)} was passed'
        logger.error(msg)
        raise TypeError(msg)


def verify_repository(client, repository=None):
    """
    Do :py:meth:`~.elasticsearch.snapshot.verify_repository` call. If it fails, raise a
    :py:exc:`~.curator.exceptions.RepositoryException`.

    :param client: A client connection object
    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :param repository: A repository name

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type repository: str

    :rtype: None
    """
    logger = logging.getLogger(__name__)
    try:
        nodes = client.snapshot.verify_repository(name=repository)['nodes']
        logger.debug('All nodes can write to the repository')
        logger.debug('Nodes with verified repository access: %s', nodes)
    except Exception as err:
        try:
            if err.status_code == 404:
                msg = (
                    f'--- Repository "{repository}" not found. Error: '
                    f'{err.meta.status}, {err.error}'
                )
            else:
                msg = (
                    f'--- Got a {err.meta.status} response from Elasticsearch.  '
                    f'Error message: {err.error}'
                )
        except AttributeError:
            msg = f'--- Error message: {err}'.format()
        report = f'Failed to verify all nodes have repository access: {msg}'
        raise RepositoryException(report) from err


def verify_snapshot_list(test):
    """
    :param test: The variable or object to test

    :type test: :py:class:`~.curator.SnapshotList`

    :returns: ``None`` if ``test`` is a proper
        :py:class:`~.curator.snapshotlist.SnapshotList` object, else raise a
        :py:class:`TypeError` exception.
    :rtype: None
    """
    # It breaks if this import isn't local to this function:
    # ImportError: cannot import name 'SnapshotList' from partially initialized module
    # 'curator.snapshotlist' (most likely due to a circular import)
    # pylint: disable=import-outside-toplevel
    from curator.snapshotlist import SnapshotList

    logger = logging.getLogger(__name__)
    if not isinstance(test, SnapshotList):
        msg = f'Not a valid SnapshotList object. Type: {type(test)} was passed'
        logger.error(msg)
        raise TypeError(msg)



"""Utility functions that get things"""

import logging
from elasticsearch8 import exceptions as es8exc
from curator.defaults.settings import EXCLUDE_SYSTEM
from curator.exceptions import (
    ConfigurationError,
    CuratorException,
    FailedExecution,
    MissingArgument,
)


def byte_size(num, suffix='B'):
    """
    :param num: The number of byte
    :param suffix: An arbitrary suffix, like ``Bytes``

    :type num: int
    :type suffix: str

    :returns: A formatted string indicating the size in bytes, with the proper unit,
        e.g. KB, MB, GB, TB, etc.
    :rtype: float
    """
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return f'{num:3.1f}{unit}{suffix}'
        num /= 1024.0
    return f'{num:.1f}Y{suffix}'


def escape_dots(stringval):
    """
    Escape any dots (periods) in ``stringval``.

    Primarily used for ``filter_path`` where dots are indicators of path nesting

    :param stringval: A string, ostensibly an index name

    :type stringval: str

    :returns: ``stringval``, but with any periods escaped with a backslash
    :retval: str
    """
    return stringval.replace('.', r'\.')


def get_alias_actions(oldidx, newidx, aliases):
    """
    :param oldidx: The old index name
    :param newidx: The new index name
    :param aliases: The aliases

    :type oldidx: str
    :type newidx: str
    :type aliases: dict

    :returns: A list of actions suitable for
        :py:meth:`~.elasticsearch.client.IndicesClient.update_aliases` ``actions``
        kwarg.
    :rtype: list
    """
    actions = []
    for alias in aliases.keys():
        actions.append({'remove': {'index': oldidx, 'alias': alias}})
        actions.append({'add': {'index': newidx, 'alias': alias}})
    return actions


def get_data_tiers(client):
    """
    Get all valid data tiers from the node roles of each node in the cluster by
    polling each node

    :param client: A client connection object
    :type client: :py:class:`~.elasticsearch.Elasticsearch`

    :returns: The available data tiers in ``tier: bool`` form.
    :rtype: dict
    """

    def role_check(role, node_info):
        if role in node_info['roles']:
            return True
        return False

    info = client.nodes.info()['nodes']
    retval = {
        'data_hot': False,
        'data_warm': False,
        'data_cold': False,
        'data_frozen': False,
    }
    for node in info:
        for role in ['data_hot', 'data_warm', 'data_cold', 'data_frozen']:
            # This guarantees we don't overwrite a True with a False.
            # We only add True values
            if role_check(role, info[node]):
                retval[role] = True
    return retval


def get_indices(client, search_pattern='*'):
    """
    Calls :py:meth:`~.elasticsearch.client.CatClient.indices`

    :param client: A client connection object
    :type client: :py:class:`~.elasticsearch.Elasticsearch`

    :returns: The current list of indices from the cluster
    :rtype: list
    """
    logger = logging.getLogger(__name__)
    indices = []
    try:
        # Doing this in two stages because IndexList also calls for these args,
        # and the unit tests need to Mock this call the same exact way.
        resp = client.cat.indices(
            index=search_pattern + ',' + EXCLUDE_SYSTEM,
            expand_wildcards='open,closed',
            h='index,status',
            format='json',
        )
    except Exception as err:
        raise FailedExecution(f'Failed to get indices. Error: {err}') from err
    if not resp:
        return indices
    for entry in resp:
        indices.append(entry['index'])
    logger.debug('All indices: %s', indices)
    return indices


def get_repository(client, repository=''):
    """
    Calls :py:meth:`~.elasticsearch.client.SnapshotClient.get_repository`

    :param client: A client connection object
    :param repository: The Elasticsearch snapshot repository to use

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type repository: str

    :returns: Configuration information for ``repository``.
    :rtype: dict
    """
    try:
        return client.snapshot.get_repository(name=repository)
    except (es8exc.TransportError, es8exc.NotFoundError) as err:
        msg = (
            f'Unable to get repository {repository}.  Error: {err} Check Elasticsearch '
            f'logs for more information.'
        )
        raise CuratorException(msg) from err


def get_snapshot(client, repository=None, snapshot=''):
    """
    Calls :py:meth:`~.elasticsearch.client.SnapshotClient.get`

    :param client: A client connection object
    :param repository: The Elasticsearch snapshot repository to use
    :param snapshot: The snapshot name, or a comma-separated list of snapshots

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type repository: str
    :type snapshot: str

    :returns: Information about the provided ``snapshot``, a snapshot (or a
        comma-separated list of snapshots). If no snapshot specified, it will
        collect info for all snapshots.  If none exist, an empty :py:class:`dict`
        will be returned.
    :rtype: dict
    """
    if not repository:
        raise MissingArgument('No value for "repository" provided')
    snapname = '*' if snapshot == '' else snapshot
    try:
        return client.snapshot.get(repository=repository, snapshot=snapshot)
    except (es8exc.TransportError, es8exc.NotFoundError) as err:
        msg = (
            f'Unable to get information about snapshot {snapname} from repository: '
            f'{repository}.  Error: {err}'
        )
        raise FailedExecution(msg) from err


def get_snapshot_data(client, repository=None):
    """
    Get all snapshots from repository and return a list.
    Calls :py:meth:`~.elasticsearch.client.SnapshotClient.get`

    :param client: A client connection object
    :param repository: The Elasticsearch snapshot repository to use

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type repository: str

    :returns: The list of all snapshots from ``repository``
    :rtype: list
    """
    if not repository:
        raise MissingArgument('No value for "repository" provided')
    try:
        return client.snapshot.get(repository=repository, snapshot="*")['snapshots']
    except (es8exc.TransportError, es8exc.NotFoundError) as err:
        msg = (
            f'Unable to get snapshot information from repository: '
            f'{repository}. Error: {err}'
        )
        raise FailedExecution(msg) from err


def get_tier_preference(client, target_tier='data_frozen'):
    """Do the tier preference thing in reverse order from coldest to hottest
    Based on the value of ``target_tier``, build out the list to use.

    :param client: A client connection object
    :param target_tier: The target data tier, e.g. data_warm.

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type target_tier: str

    :returns: A suitable tier preference string in csv format
    :rtype: str
    """
    tiermap = {
        'data_content': 0,
        'data_hot': 1,
        'data_warm': 2,
        'data_cold': 3,
        'data_frozen': 4,
    }
    tiers = get_data_tiers(client)
    test_list = []
    for tier in ['data_hot', 'data_warm', 'data_cold', 'data_frozen']:
        if tier in tiers and tiermap[tier] <= tiermap[target_tier]:
            test_list.insert(0, tier)
    if target_tier == 'data_frozen':
        # We're migrating to frozen here. If a frozen tier exists, frozen searchable
        # snapshot mounts should only ever go to the frozen tier.
        if 'data_frozen' in tiers and tiers['data_frozen']:
            return 'data_frozen'
    # If there are no  nodes with the 'data_frozen' role...
    preflist = []
    for key in test_list:
        # This ordering ensures that colder tiers are prioritized
        if key in tiers and tiers[key]:
            preflist.append(key)
    # If all of these are false, then we have no data tiers and must use 'data_content'
    if not preflist:
        return 'data_content'
    # This will join from coldest to hottest as csv string,
    # e.g. 'data_cold,data_warm,data_hot'
    return ','.join(preflist)


def get_write_index(client, alias):
    """
    Calls :py:meth:`~.elasticsearch.client.IndicesClient.get_alias`

    :param client: A client connection object
    :param alias: An alias name

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type alias: str

    :returns: The the index name associated with the alias that is designated
        ``is_write_index``
    :rtype: str
    """
    try:
        response = client.indices.get_alias(index=alias)
    except Exception as exc:
        raise CuratorException(f'Alias {alias} not found') from exc
    # If there are more than one in the list, one needs to be the write index
    # otherwise the alias is a one to many, and can't do rollover.
    retval = None
    if len(list(response.keys())) > 1:
        for index in list(response.keys()):
            try:
                if response[index]['aliases'][alias]['is_write_index']:
                    retval = index
            except KeyError as exc:
                raise FailedExecution(
                    'Invalid alias: is_write_index not found in 1 to many alias'
                ) from exc
    else:
        # There's only one, so this is it
        retval = list(response.keys())[0]
    return retval


def index_size(client, idx, value='total'):
    """
    Calls :py:meth:`~.elasticsearch.client.IndicesClient.stats`

    :param client: A client connection object
    :param idx: An index name
    :param value: One of either ``primaries`` or ``total``

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type idx: str
    :type value: str

    :returns: The sum of either ``primaries`` or ``total`` shards for index ``idx``
    :rtype: integer
    """
    fpath = f'indices.{escape_dots(idx)}.{value}.store.size_in_bytes'
    return client.indices.stats(index=idx, filter_path=fpath)['indices'][idx][value][
        'store'
    ]['size_in_bytes']


def meta_getter(client, idx, get=None):
    """Meta Getter
    Calls :py:meth:`~.elasticsearch.client.IndicesClient.get_settings` or
    :py:meth:`~.elasticsearch.client.IndicesClient.get_alias`

    :param client: A client connection object
    :param idx: An Elasticsearch index
    :param get: The kind of get to perform, e.g. settings or alias

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type idx: str
    :type get: str

    :returns: The settings from the get call to the named index
    :rtype: dict
    """
    logger = logging.getLogger(__name__)
    acceptable = ['settings', 'alias']
    if not get:
        raise ConfigurationError('"get" can not be a NoneType')
    if get not in acceptable:
        raise ConfigurationError(f'"get" must be one of {acceptable}')
    retval = {}
    try:
        if get == 'settings':
            retval = client.indices.get_settings(index=idx)[idx]['settings']['index']
        elif get == 'alias':
            retval = client.indices.get_alias(index=idx)[idx]['aliases']
    except es8exc.NotFoundError as missing:
        logger.error('Index %s was not found!', idx)
        raise es8exc.NotFoundError from missing
    except KeyError as err:
        logger.error('Key not found: %s', err)
        raise KeyError from err
    # pylint: disable=broad-except
    except Exception as exc:
        logger.error('Exception encountered: %s', exc)
    return retval


def name_to_node_id(client, name):
    """
    Calls :py:meth:`~.elasticsearch.client.NodesClient.info`

    :param client: A client connection object
    :param name: The node ``name``

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type name: str

    :returns: The node_id of the node identified by ``name``
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    fpath = 'nodes'
    info = client.nodes.info(filter_path=fpath)
    for node in info['nodes']:
        if info['nodes'][node]['name'] == name:
            logger.debug('Found node_id "%s" for name "%s".', node, name)
            return node
    logger.error('No node_id found matching name: "%s"', name)
    return None


def node_id_to_name(client, node_id):
    """
    Calls :py:meth:`~.elasticsearch.client.NodesClient.info`

    :param client: A client connection object
    :param node_id: The node ``node_id``

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type node_id: str

    :returns: The name of the node identified by ``node_id``
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    fpath = f'nodes.{node_id}.name'
    info = client.nodes.info(filter_path=fpath)
    name = None
    if node_id in info['nodes']:
        name = info['nodes'][node_id]['name']
    else:
        logger.error('No node_id found matching: "%s"', node_id)
    logger.debug('Name associated with node_id "%s": %s', node_id, name)
    return name


def node_roles(client, node_id):
    """
    Calls :py:meth:`~.elasticsearch.client.NodesClient.info`

    :param client: A client connection object
    :param node_id: The node ``node_id``

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type node_id: str

    :returns: The list of roles assigned to the node identified by ``node_id``
    :rtype: list
    """
    fpath = f'nodes.{node_id}.roles'
    return client.nodes.info(filter_path=fpath)['nodes'][node_id]['roles']


def single_data_path(client, node_id):
    """
    In order for a shrink to work, it should be on a single filesystem, as shards
    cannot span filesystems. Calls :py:meth:`~.elasticsearch.client.NodesClient.stats`

    :param client: A client connection object
    :param node_id: The node ``node_id``

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type node_id: str

    :returns: ``True`` if the node has a single filesystem, else ``False``
    :rtype: bool
    """
    fpath = f'nodes.{node_id}.fs.data'
    response = client.nodes.stats(filter_path=fpath)
    return len(response['nodes'][node_id]['fs']['data']) == 1



"""Index replica count action class"""

import logging
from curator.exceptions import MissingArgument
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import chunk_index_list, report_failure, show_dry_run, to_csv
from curator.helpers.waiters import wait_for_it


class Replicas:
    """Replica Action Class"""

    def __init__(
        self, ilo, count=None, wait_for_completion=False, wait_interval=9, max_wait=-1
    ):
        """
        :param ilo: An IndexList Object
        :param count: The count of replicas per shard
        :param wait_for_completion: Wait for completion before returning.
        :param wait_interval: Seconds to wait between completion checks.
        :param max_wait: Maximum number of seconds to ``wait_for_completion``

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type count: int
        :type wait_for_completion: bool
        :type wait_interval: int
        :type max_wait: int
        """
        verify_index_list(ilo)
        # It's okay for count to be zero
        if count == 0:
            pass
        elif not count:
            raise MissingArgument('Missing value for "count"')

        #: The :py:class:`~.curator.indexlist.IndexList` object passed from
        #: param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: Object attribute that gets the value of param ``count``.
        self.count = count
        #: Object attribute that gets the value of param ``wait_for_completion``.
        self.wfc = wait_for_completion
        #: Object attribute that gets the value of param ``wait_interval``.
        self.wait_interval = wait_interval
        #: Object attribute that gets the value of param ``max_wait``.
        self.max_wait = max_wait
        self.loggit = logging.getLogger('curator.actions.replicas')

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        show_dry_run(self.index_list, 'replicas', count=self.count)

    def do_action(self):
        """
        Update ``number_of_replicas`` with :py:attr:`count` and
        :py:meth:`~.elasticsearch.client.IndicesClient.put_settings` to indices in
        :py:attr:`index_list`
        """
        self.loggit.debug(
            'Cannot get update replica count of closed indices. Omitting any '
            'closed indices.'
        )
        self.index_list.filter_closed()
        self.index_list.empty_list_check()
        msg = (
            f'Setting the replica count to {self.count} for '
            f'{len(self.index_list.indices)} indices: {self.index_list.indices}'
        )
        self.loggit.info(msg)
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                self.client.indices.put_settings(
                    index=to_csv(lst), settings={'number_of_replicas': self.count}
                )
                if self.wfc and self.count > 0:
                    msg = (
                        f'Waiting for shards to complete replication for indices: '
                        f'{to_csv(lst)}'
                    )
                    self.loggit.debug(msg)
                    wait_for_it(
                        self.client,
                        'replicas',
                        wait_interval=self.wait_interval,
                        max_wait=self.max_wait,
                    )
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Filter element schema definitions

All member functions return a :class:`voluptuous.schema_builder.Schema` object
"""

from voluptuous import All, Any, Boolean, Coerce, Optional, Range, Required
from curator.defaults import settings

# pylint: disable=unused-argument, line-too-long


def aliases(**kwargs):
    """
    :returns: {Required('aliases'): Any(list, str)}
    """
    return {Required('aliases'): Any(list, str)}


def allocation_type(**kwargs):
    """
    :returns: {Optional('allocation_type', default='require'):
        All(Any(str), Any('require', 'include', 'exclude'))}
    """
    return {
        Optional('allocation_type', default='require'): All(
            Any(str), Any('require', 'include', 'exclude')
        )
    }


def count(**kwargs):
    """
    This setting is only used with the count filtertype and is required

    :returns: {Required('count'): All(Coerce(int), Range(min=1))}
    """
    return {Required('count'): All(Coerce(int), Range(min=1))}


def date_from(**kwargs):
    """
    This setting is only used with the period filtertype.

    :returns: {Optional('date_from'): Any(str)}
    """
    return {Optional('date_from'): Any(str)}


def date_from_format(**kwargs):
    """
    This setting is only used with the period filtertype.

    :returns: {Optional('date_from_format'): Any(str)}
    """
    return {Optional('date_from_format'): Any(str)}


def date_to(**kwargs):
    """
    This setting is only used with the period filtertype.

    :returns: {Optional('date_to'): Any(str)}
    """
    return {Optional('date_to'): Any(str)}


def date_to_format(**kwargs):
    """
    This setting is only used with the period filtertype.

    :returns: {Optional('date_to_format'): Any(str)}
    """
    return {Optional('date_to_format'): Any(str)}


def direction(**kwargs):
    """
    This setting is only used with the ``age`` filtertype.

    :returns: {Required('direction'): Any('older', 'younger')}
    """
    return {Required('direction'): Any('older', 'younger')}


def disk_space(**kwargs):
    """
    This setting is only used with the ``space`` filtertype and is required

    :returns: {Required('disk_space'): Any(Coerce(float))}
    """
    return {Required('disk_space'): Any(Coerce(float))}


def epoch(**kwargs):
    """
    This setting is only used with the ``age`` filtertype.

    :returns: {Optional('epoch', default=None): Any(Coerce(int), None)}
    """

    return {Optional('epoch', default=None): Any(Coerce(int), None)}


def exclude(**kwargs):
    """
    This setting is available in all filter types.
    The default ``val`` is ``True`` if ``exclude`` in ``kwargs``, otherwise ``False``

    :returns: {Optional('exclude', default=val):
        Any(bool, All(Any(str), Boolean()))}
    """
    val = bool('exclude' in kwargs and kwargs['exclude'])
    # pylint: disable=no-value-for-parameter
    return {Optional('exclude', default=val): Any(bool, All(Any(str), Boolean()))}


def field(**kwargs):
    """
    This setting is only used with the ``age`` filtertype.

    :returns: {Required('field'): Any(str)} if ``kwargs['required']`` is ``True``
        otherwise {Optional('field'): Any(str)}
    """
    if 'required' in kwargs and kwargs['required']:
        return {Required('field'): Any(str)}
    return {Optional('field'): Any(str)}


def intersect(**kwargs):
    """
    This setting is only used with the period filtertype when using field_stats, i.e.
    indices only.

    :returns: {Optional('intersect', default=False):
        Any(bool, All(Any(str), Boolean()))}
    """
    # pylint: disable=no-value-for-parameter
    return {Optional('intersect', default=False): Any(bool, All(Any(str), Boolean()))}


def key(**kwargs):
    """
    This setting is only used with the allocated filtertype.

    :returns: {Required('key'): Any(str)}
    """
    return {Required('key'): Any(str)}


def kind(**kwargs):
    """
    This setting is only used with the pattern filtertype and is required

    :returns: {Required('kind'): Any('prefix', 'suffix', 'timestring', 'regex')}
    """
    return {Required('kind'): Any('prefix', 'suffix', 'timestring', 'regex')}


def max_num_segments(**kwargs):
    """
    :returns: {Required('max_num_segments'): All(Coerce(int), Range(min=1))}
    """
    return {Required('max_num_segments'): All(Coerce(int), Range(min=1))}


def number_of_shards(**kwargs):
    """
    :returns: {Required('number_of_shards'): All(Coerce(int), Range(min=1))}
    """
    return {Required('number_of_shards'): All(Coerce(int), Range(min=1))}


def pattern(**kwargs):
    """
    :returns: {Optional('pattern'): Any(str)}
    """
    return {Optional('pattern'): Any(str)}


def period_type(**kwargs):
    """
    This setting is only used with the period filtertype.

    :returns: {Optional('period_type', default='relative'):
        Any('relative', 'absolute')}
    """
    return {Optional('period_type', default='relative'): Any('relative', 'absolute')}


def range_from(**kwargs):
    """
    :returns: {Optional('range_from'): Coerce(int)}
    """
    return {Optional('range_from'): Coerce(int)}


def range_to(**kwargs):
    """
    :returns: {Optional('range_to'): Coerce(int)}
    """
    return {Optional('range_to'): Coerce(int)}


def reverse(**kwargs):
    """
    Only used with ``space`` filtertype. Should be ignored if ```use_age``` is True

    :returns: {Optional('reverse', default=True):
        Any(bool, All(Any(str), Boolean()))}
    """
    # pylint: disable=no-value-for-parameter
    return {Optional('reverse', default=True): Any(bool, All(Any(str), Boolean()))}


def shard_filter_behavior(**kwargs):
    """
    This setting is only used with the shards filtertype and defaults to 'greater_than'.

    :returns: {Optional('shard_filter_behavior', default='greater_than'):
        Any('greater_than', 'less_than', 'greater_than_or_equal',
        'less_than_or_equal', 'equal')}
    """
    return {
        Optional('shard_filter_behavior', default='greater_than'): Any(
            'greater_than',
            'less_than',
            'greater_than_or_equal',
            'less_than_or_equal',
            'equal',
        )
    }


def size_threshold(**kwargs):
    """
    This setting is only used with the size filtertype and is required

    :returns: {Required('size_threshold'): Any(Coerce(float))}
    """
    return {Required('size_threshold'): Any(Coerce(float))}


def size_behavior(**kwargs):
    """
    This setting is only used with the size filtertype and defaults to 'primary'.

    :returns: {Optional('size_behavior', default='primary'):
        Any('primary', 'total')}
    """
    return {Optional('size_behavior', default='primary'): Any('primary', 'total')}


def source(**kwargs):
    """
    This setting is only used with the ``age`` filtertype, or with the ``space``
    filtertype when ``use_age`` is set to True.

    :ivar valuelist: If ``kwargs['action']`` is in
        :py:func:`curator.defaults.settings.snapshot_actions`, then it is
        ``Any('name', 'creation_date')``, otherwise
        ``Any('name', 'creation_date', 'field_stats')``
    :returns: {Required('source'): valuelist} if ``kwargs['required']``, else
        {Optional('source'): valuelist}
    """
    if 'action' in kwargs and kwargs['action'] in settings.snapshot_actions():
        valuelist = Any('name', 'creation_date')
    valuelist = Any('name', 'creation_date', 'field_stats')
    if 'required' in kwargs and kwargs['required']:
        return {Required('source'): valuelist}
    return {Optional('source'): valuelist}


def state(**kwargs):
    """
    This setting is only used with the state filtertype.

    :returns: {Optional('state', default='SUCCESS'):
        Any('SUCCESS', 'PARTIAL', 'FAILED', 'IN_PROGRESS')}
    """
    return {
        Optional('state', default='SUCCESS'): Any(
            'SUCCESS', 'PARTIAL', 'FAILED', 'IN_PROGRESS'
        )
    }


def stats_result(**kwargs):
    """
    This setting is only used with the ``age`` filtertype.

    :returns: {Optional('stats_result', default='min_value'):
        Any('min_value', 'max_value')}
    """
    return {
        Optional('stats_result', default='min_value'): Any('min_value', 'max_value')
    }


def timestring(**kwargs):
    """
    This setting is only used with the ``age`` filtertype, or with the ``space``
    filtertype if ``use_age`` is set to ``True``.

    :returns: {Required('timestring'): Any(str)} if ``kwargs['required']`` else
      {Optional('timestring', default=None): Any(None, str)}
    """
    if 'required' in kwargs and kwargs['required']:
        return {Required('timestring'): Any(str)}
    return {Optional('timestring', default=None): Any(None, str)}


def threshold_behavior(**kwargs):
    """
    This setting is only used with the space and size filtertype and defaults to
    'greater_than'.

    :returns: {Optional('threshold_behavior', default='greater_than'):
        Any('greater_than', 'less_than')}
    """
    return {
        Optional('threshold_behavior', default='greater_than'): Any(
            'greater_than', 'less_than'
        )
    }


def unit(**kwargs):
    """
    This setting is only used with the ``age`` filtertype, or with the ``space``
    filtertype if ``use_age`` is set to ``True``.

    :returns: {Required('unit'): Any('seconds', 'minutes', 'hours', 'days', 'weeks',
        'months', 'years')}
    """
    return {
        Required('unit'): Any(
            'seconds', 'minutes', 'hours', 'days', 'weeks', 'months', 'years'
        )
    }


def unit_count(**kwargs):
    """
    This setting is only used with the ``age`` filtertype, or with the ``space``
    filtertype if ``use_age`` is set to ``True``.

    :returns: {Required('unit_count'): Coerce(int)}
    """
    return {Required('unit_count'): Coerce(int)}


def unit_count_pattern(**kwargs):
    """
    This setting is used with the ``age`` filtertype to define whether the
    ``unit_count`` value is taken from the configuration or read from the index
    name via a regular expression

    :returns: {Optional('unit_count_pattern'): Any(str)}
    """
    return {Optional('unit_count_pattern'): Any(str)}


def use_age(**kwargs):
    """
    Use of this setting requires the additional setting, ``source``.

    :returns: {Optional('use_age', default=False):
        Any(bool, All(Any(str), Boolean()))}
    """
    # pylint: disable=no-value-for-parameter
    return {Optional('use_age', default=False): Any(bool, All(Any(str), Boolean()))}


def value(**kwargs):
    """
    This setting is only used with the ``pattern`` filtertype and is a required
    setting. There is a separate value option associated with the ``Allocation``
    action, and the ``allocated`` filtertype.

    :returns: {Required('value'): Any(str)}
    """
    return {Required('value'): Any(str)}


def week_starts_on(**kwargs):
    """
    :returns: {Optional('week_starts_on', default='sunday'):
        Any('Sunday', 'sunday', 'SUNDAY', 'Monday', 'monday', 'MONDAY', None)}
    """
    return {
        Optional('week_starts_on', default='sunday'): Any(
            'Sunday', 'sunday', 'SUNDAY', 'Monday', 'monday', 'MONDAY', None
        )
    }
